#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define SUCCESS												0
#define ERR_INVALID_HANDLE						-1
#define ERR_NULL_POINTER							-2
#define ERR_NULL_DELIMITER						-3
#define ERR_PARSE_ERROR								-4
#define ERR_NEGATIVE_TOKEN						-5
#define ERR_TOKEN_OUT_OF_RANGE				-6

char str[766];

#if defined (__MACH__)
#include <ctype.h>
#endif

#if defined (WIN32)
#include <errno.h>
#include <assert.h>

/* Always add at least this many bytes when extending the buffer.  */
#define MIN_CHUNK 64

/* Read up to (and including) a TERMINATOR from STREAM into *LINEPTR
   + OFFSET (and null-terminate it). *LINEPTR is a pointer returned from
   malloc (or NULL), pointing to *N characters of space.  It is realloc'd
   as necessary.  Return the number of characters read (not including the
   null terminator), or -1 on error or EOF.  On a -1 return, the caller
   should check feof(), if not then errno has been set to indicate
   the error.  */

int getstr(char **lineptr, size_t *n, FILE *stream, char terminator, int offset) {

  int nchars_avail;		/* Allocated but unused chars in *LINEPTR.  */
  char *read_pos;			/* Where we're reading into *LINEPTR. */
  int ret;
	int save_errno;
	register int c;

  if (!lineptr || !n || !stream) {
		errno = EINVAL;
		return -1;
	}
  if (!*lineptr) {
		*n = MIN_CHUNK;
		*lineptr = malloc (*n);
		if (!*lineptr) {
	  	errno = ENOMEM;
	  	return -1;
		}
  }
  nchars_avail = *n - offset;
  read_pos = *lineptr + offset;
  for (;;) {
		c = getc (stream);
		save_errno = errno;
		/* We always want at least one char left in the buffer, since we
	 	always (unless we get an error while reading the first char)
	 	NUL-terminate the line buffer.  */
    assert((*lineptr + *n) == (read_pos + nchars_avail));
    if (nchars_avail < 2) {
	  	if (*n > MIN_CHUNK)
	    	*n *= 2;
	  	else
	    	*n += MIN_CHUNK;
	  	nchars_avail = *n + *lineptr - read_pos;
	  	*lineptr = realloc (*lineptr, *n);
	  	if (!*lineptr) {
      	errno = ENOMEM;
	      return -1;
	    }
	  	read_pos = *n - nchars_avail + *lineptr;
	  	assert((*lineptr + *n) == (read_pos + nchars_avail));
		}
    if (ferror (stream)) {
	  	/* Might like to return partial line, but there is no
     	place for us to store errno.  And we don't want to just
     	lose errno.  */
	  	errno = save_errno;
	  	return -1;
		}
		if (c == EOF) {
	  	/* Return partial line, if any.  */
	  	if (read_pos == *lineptr)
	    	return -1;
	  	else
	    	break;
		}
    *read_pos++ = c;
    nchars_avail--;
		if (c == terminator)
		/* Return the line.  */
			break;
  }
  /* Done - NUL terminate and return the number of chars read.  */
  *read_pos = '\0';
  ret = read_pos - (*lineptr + offset);
  return ret;
}

int mygetline(char **lineptr, size_t *n, FILE *stream) {

  return getstr(lineptr, n, stream, '\n', 0);
}
#endif

void *get_peano_representation(unsigned char arabic_number) {

	unsigned char i;
	char *ptr;
	
	ptr = str;																			
	for (i=0; i<arabic_number; i++) {
		strcat(ptr, "s(");
		ptr += 2;
	}
	*ptr = '0';
	ptr++;
	for (i=0; i<arabic_number; i++) {
		*ptr = ')';
		ptr++;
	}
	*ptr = '\0';
	ptr++;
}

int parse_line(char *line, char *delim, FILE *fout) {
	
	long int parsed_values[4];
	int tokens_found = 0, delim_size, i;
	char *startptr, *endptr;

	if (fout == NULL)
		return ERR_INVALID_HANDLE;																											// Reject invalid handle for the output file.
	if (!line || !delim)
		return ERR_NULL_POINTER;																												// Reject NULL pointers for line and/or delim.
	if ((delim_size = strlen(delim)) == 0)  
		return ERR_NULL_DELIMITER;  																										// Delimiter not provided.
	startptr = line;
	while (1) {																																				// Infinite loop.
		parsed_values[tokens_found] = strtol(startptr, &endptr, 10);										// Parse the next value.
		if (startptr == endptr)
			return ERR_PARSE_ERROR;																												// Could not parse any valid value.
		if (parsed_values[tokens_found] < 0)
			return ERR_NEGATIVE_TOKEN;																										// Illegal (negative) value was parsed.
		if (parsed_values[tokens_found] > 255)																					// Illegal (greater than 255) value was parsed.
			return ERR_TOKEN_OUT_OF_RANGE;								
		tokens_found++;																																	// Increment the token counter.
		if (tokens_found == 4) {																												// Check whether have we already found 4 tokens.
			startptr = endptr;																														// Update the startptr pointer to point to the character right after the parsed int value.			
			while (isspace(*startptr) && *startptr != '\n')																// If so, make sure that all remaining characters in the line are blanks.
				startptr++;
			if (*startptr == '\n')
				break;
			else
				return ERR_PARSE_ERROR;																													
		}
		startptr = endptr;																															// Update the startptr pointer to point to the character right after the parsed int value.
		while (isspace(*startptr) && *startptr != '\n')																	// Discard all blanks until we find a non-blank character.
			startptr++;
		if (!strncmp(startptr, delim, delim_size))
			startptr += delim_size;																												// A delimiter was found as expected. Update the startptr pointer to point the character after the delimiter.
		else
			return ERR_PARSE_ERROR;																												// A delimiter was not found as expected. Return parse error.
	}
	fputs("lit(", fout);																															// Write the functor plus the opening bracket to the file.
	for (i=0; i<4; i++) {																															// For each parsed value:
		if (parsed_values[i] == 0)																											// If the current parse value is zero, then write a '0' (its Peano representation) to the output file.	
			fputs("0", fout);
		else {																																					// Otherwise allocate memory to create a placeholder for the Peano representation of the current value.
			str[0] = '\0';																																// Nullify the first character of the Peano's placeholder to allow for strcat to work properly.
			get_peano_representation((unsigned char)parsed_values[i]);										// Get the Peano representation for the current value.
			fputs(str, fout);																															// Write the Peano string to the output file.
		}
		if (i != 3)																																			// If this is not the last value, print the delimiter.
			fputs(delim, fout);
		else
			fputs(").\n", fout);																													// Print the closing bracket plus the clause termination character plus a line feed character to the file otherwise.
	}

	return SUCCESS;																																		// Return control to the caller.
}

void usage(char *progname) {
	fprintf(stderr, "Usage: %s <input_file> <output_file> [--discard-junk]\n", progname);
	return;
}

int main(int argc, char *argv[]) {

	char delim[] = ",";
	int i, ret, lineno = 1;
	FILE *fin, *fout;
	char message[256];
	ssize_t read;
	ssize_t len;
	char *line = NULL;
	int discard_junk = 0;

	fprintf(stdout, "Facts base generator for the number detection assignment.\n");
	fprintf(stdout, "(c) 2015 Miguel Garcia Remesal.\n");
	if ((argc < 3) || (argc > 4)) {
		usage(argv[0]);
		exit(-1);
	}
	if (argc == 4) {
		if (strcmp(argv[3], "--discard-junk")) {
			usage(argv[0]);
			exit(-1);
		}
		else
			discard_junk = 1;
	}
	fin = fopen(argv[1], "r");
	if (!fin) {
		sprintf(message, "fopen: %s", argv[1]);
		perror(message);
		exit(-1); 
	}
	fout = fopen(argv[2], "w+");
	if (!fout) {
		sprintf(message, "fopen: %s", argv[2]);
		perror(message);
		exit(-1);
	}
#if defined (WIN32)
	while ((read = mygetline(&line, &len, fin)) != -1) {
#else
	while ((read = getline(&line, &len, fin)) != -1) {
#endif
		if (strlen(line) == 1) {
			if (!discard_junk)
				fputs("\n", fout);
			lineno++;
			continue;
		}
		if (*line == '%') {
			if (!discard_junk) {
				fputs(line, fout);
			}
			lineno++;
			continue;
		}
		ret = parse_line(line, delim, fout);
		if (ret != SUCCESS) {
			switch (ret) {
				case ERR_INVALID_HANDLE:
					fprintf(stderr, "parse_line: invalid handle provided for output file %s.", argv[2]);
					break;
				case ERR_NULL_POINTER:
					fprintf(stderr, "parse_line: NULL pointer exception (for either the line, the delimiter or both).\n");
					break;
				case ERR_NULL_DELIMITER:
					fprintf(stderr, "parse_line: the delimiter cannot be an empty.\n");
					break;
				case ERR_PARSE_ERROR:
					fprintf(stderr, "parse_line: parse error at line %d.\n", lineno);
					break;
				case ERR_NEGATIVE_TOKEN:
					fprintf(stderr, "parse_line: invalid (negative) token found at line %d.\n", lineno);
					break;
				case ERR_TOKEN_OUT_OF_RANGE:
					fprintf(stderr, "parse_line: token out of range (>255) found at line %d.\n", lineno);
					break;
				default:
					fprintf(stderr, "parse_line: unexpected error.\n");
					break;
			}
			exit(-1);
		}
		lineno++;
	}
	fprintf(stdout, "The output file was successfully generated.\n");
	fclose(fin);
	fclose(fout);	
	return 0;
}
